package ap.proof.goal;

import ap.util.{Debug, Logic, APTestCase, PlainRange}
import ap.terfor._
import ap.terfor.substitutions.ConstantSubst
import ap.terfor.equations.{EquationConj, NegEquationConj}
import ap.terfor.linearcombination.LinearCombination
import ap.basetypes.IdealInt

class TestEquationalProving(n : String) extends APTestCase(n) {

  def runTest = {
    n match {
      case "testEqs1" => testEqs1
      case "testEqs2" => testEqs2
      case "testQuantifiedEqs1" => testQuantifiedEqs1
      case "testDivisibility1" => testDivisibility1
    }
  }

  private val consts = for (i <- Array.range(0, 10)) yield new ConstantTerm("c" + i)
  private val constsOne = consts ++ List(OneTerm)
  private val to = (TermOrder.EMPTY /: consts)((o, c) => o.extend(c, Set.empty))
  
  private def randomInput(len : Int) = for (i <- PlainRange(0, len))
                                       yield (IdealInt(Debug.random(-20, 20)),
                                              constsOne(Debug.random(0, constsOne.size)))

  private def randomLC(len : Int) = LinearCombination(randomInput(len), to)

  private def randomAC(eqNum : Int, maxEqSize : Int) : ArithConj = {
    val posEqNum = Debug.random(0, eqNum+1)
    randomAC(posEqNum, eqNum - posEqNum, maxEqSize)
  }

  private def randomAC(posEqNum : Int, negEqNum : Int, maxEqSize : Int) : ArithConj = {
    val inputConj = (for (len <- Debug.randoms(0, maxEqSize+1)) yield randomLC(len))
                    .take(posEqNum).toList
    val eqConj = EquationConj(inputConj, to)
    val negInputConj = (for (len <- Debug.randoms(0, maxEqSize+1)) yield randomLC(len))
                       .take(negEqNum).toList
    val negEqConj = NegEquationConj(negInputConj, to)
    
    ArithConj(eqConj, negEqConj, to)
  }

  def testEqs1 = {
    for (eqNum <- PlainRange(0, 15); eqSize <- PlainRange(0, 15)) {
      val ac = randomAC(eqNum, eqSize)
      val facts = Conjunction.conj(Array(ac), to)
      for (elimNum <- PlainRange(0, consts.size + 1)) {
        val eliminatedConsts = Set.empty ++ consts.drop(consts.size - elimNum)
        val goal = Goal(facts, TaskManager.EMPTY, 0, eliminatedConsts, to)
      
//        println
//        println(goal)
    
        val tree = FactsNormalisationTask(goal).toTree(goal)

        assertNormalisedTree(tree)

//          println("After one step:")
//          println(tree)
      
          
      }
    }
  }
  
  def testEqs2 = {
    var eqs : Formula =
      EquationConj(LinearCombination(Array((IdealInt.ONE, consts(0)),
                                           (IdealInt(-5), consts(1)),
                                           (IdealInt(-1), OneTerm)), to), to)
    for (premNum <- PlainRange(1, consts.size - 1)) {
      eqs = Conjunction.conj(Array(EquationConj(LinearCombination(
                                     Array((IdealInt(4), consts(premNum)),
                                           (IdealInt(-5), consts(premNum+1)),
                                           (IdealInt(-1), OneTerm)), to), to),
                                   eqs), to)
      
      val goal = Goal(Conjunction.negate(eqs, to), to)
      
//        println(goal)
//        println(goal.closingConstraint)
    
        val tree = ExhaustiveProver(goal)
//          println("After many steps:")
//          println(tree)
//          println(tree.closingConstraint)
      
          assertNormalisedTree(tree)
    }
  }
  
  def testDivisibility1 = {
    val div1 = Conjunction.quantify(Array(Quantifier.EX),
      EquationConj(LinearCombination(Array((IdealInt(5), VariableTerm(0)),
                                           (IdealInt(1), consts.last),
                                           (IdealInt(1), OneTerm)), to), to), to)
    val div2 = Conjunction.quantify(Array(Quantifier.EX),
      EquationConj(LinearCombination(Array((IdealInt(5), VariableTerm(0)),
                                           (IdealInt(1), consts.last),
                                           (IdealInt(2), OneTerm)), to), to), to)
    val f = Conjunction.disj(Array(div1, div2), to)

    for (elimConsts 
         <- List(Set.empty, Set(consts.last)).asInstanceOf[List[Set[ConstantTerm]]]) {
      val goal = Goal(f, elimConsts, to)

      println(goal)
    
      val tree = ExhaustiveProver(goal)
      println("After many steps:")
      println(tree)
      
//      assertNormalisedTree(tree)
    }
  }
  
  def testQuantifiedEqs1 = {
    ConstraintSimplifier.enabled = false
    
    var num : Int = 0
    
    for (eqNum <- PlainRange(0, 15); eqSize <- PlainRange(0, 15)) {
      val ac = randomAC(eqNum, eqSize)

      for (quantNum <- List(0, 1, 2, 5, 10)) {
        var formula : Formula = ac
        for ((c, q) <- (Debug.randoms(0, consts.size) zip Debug.randoms(0, 2))
                       .take(quantNum)) {
          val const = consts(c)
          val quan = if (q == 0) Quantifier.ALL else Quantifier.EX
          val subst = new ConstantSubst(const, VariableTerm(0), to)
          formula = Conjunction.quantify(Array(quan), subst(formula), to)
        }
        
        println
        println(num)
        num = num + 1
        println(formula)
        
        val goal = Goal(formula, to)
      
        println(goal)
    
          val tree = ExhaustiveProver(goal)
          println("After many steps:")
          println(tree)
      
//          assertNormalisedTree(tree)
      }
    }

    ConstraintSimplifier.enabled = true
  }
  
  private def assertNormalisedTree(t : ProofTree) : Unit = t match {
    case t : ProofTreeOneChild => assertNormalisedTree(t.subtree)
    case t : AndTree => for (subtree <- t.subtrees.elements)
                          assertNormalisedTree(subtree)
    case goal : Goal => {
      assert(goal.tasks.isEmpty)
      if (!goal.facts.isFalse) {
        assert(Logic.forall(for (lc <- goal.facts.arithConj.positiveEqs.elements)
                            yield isNormalisedPosEq(lc, goal)))
        assert(Logic.forall(for (lc <- goal.facts.arithConj.negativeEqs.elements)
                            yield isNormalisedNegEq(lc, goal)))
      }
    }
  }
  
  private def isNormalisedPosEq(lc : LinearCombination, goal : Goal) =
    lc.leadingCoeff.isOne
/*    if (goal eliminates lc.leadingTerm)
      (lc.leadingCoeff.isOne ||
       Logic.forall(1, lc.size, (i) => !(goal eliminates (lc(i) _2))))
    else
      lc.leadingCoeff.isOne */

  private def isNormalisedNegEq(lc : LinearCombination, goal : Goal) =
    Logic.forall(for (posLc <- goal.facts.arithConj.positiveEqs.elements)
                 yield ((lc.leadingTerm != posLc.leadingTerm) &&
                        (if (goal eliminates posLc.leadingTerm)
                           lc.get(posLc.leadingTerm) isAbsMinMod posLc.leadingCoeff
                         else
                           doesNotContain(lc, posLc.leadingTerm))))

  private def doesNotContain(tf : TerFor, t : Term) : Boolean = t match {
    case t : ConstantTerm => !(tf.constants contains t)
    case t : VariableTerm => !(tf.variables contains t)
    case _ => true
  }
}
