package ap.terfor;

import ap.util.{Debug, Logic, APTestCase, PlainRange, FilterIt}
import ap.basetypes.IdealInt
import ap.terfor.linearcombination.LinearCombination
import ap.terfor.equations.{EquationConj, NegEquationConj}

class TestPropConnectives(n : String) extends APTestCase(n) {

  def runTest = {
    n match {
      case "testArithConj1" => testArithConj1
      case "testArithConj2" => testArithConj2
      case "testConj1" => testConj1
    }
  }

  private val consts = for (i <- Array.range(0, 10)) yield new ConstantTerm("c" + i)
  private val vars = for (i <- Array.range(0, 20)) yield new VariableTerm(i)
  private val constsVarsOne = consts ++ vars ++ List(OneTerm)
  private val to = (TermOrder.EMPTY /: consts)((o, c) => o.extend(c, Set.empty))
  private val toRev = (consts :\ TermOrder.EMPTY)((c, o) => o.extend(c, Set.empty))
  
  private def randomInput(len : Int) = for (i <- PlainRange(0, len))
                                       yield (IdealInt(Debug.random(-20, 20)),
                                              constsVarsOne(Debug.random(0, constsVarsOne.size)))

  private def randomLC(len : Int) = LinearCombination(randomInput(len), to)

  private def randomAC(eqNum : Int, maxEqSize : Int) : ArithConj = {
    val posEqNum = Debug.random(0, eqNum+1)
    randomAC(posEqNum, eqNum - posEqNum, maxEqSize)
  }

  private def randomAC(posEqNum : Int, negEqNum : Int, maxEqSize : Int) : ArithConj = {
    val inputConj = (for (len <- Debug.randoms(0, maxEqSize+1)) yield randomLC(len))
                    .take(posEqNum).toList
    val eqConj = EquationConj(inputConj, to)
    val negInputConj = (for (len <- Debug.randoms(0, maxEqSize+1)) yield randomLC(len))
                       .take(negEqNum).toList
    val negEqConj = NegEquationConj(negInputConj, to)
    
    ArithConj(eqConj, negEqConj, to)
  }
  
  private def randomConj(sizeFactor : Double,
                         maxEqNum : Int, maxEqSize : Int) : Conjunction = {
    val quans : List[Quantifier] = (for (i <- Debug.randoms(0, 2))
                                    yield (if (i==0) Quantifier.ALL else Quantifier.EX))
                                   .take(Debug.random(0, 10)).toList
    val arithConj = randomAC(Debug.random(0, maxEqNum),
                             Debug.random(0, maxEqSize))
    val negConjs = if (arithConj.isFalse)
                     NegatedConjunctions.TRUE
                   else
                     randomNegConjs(sizeFactor - 1, maxEqNum, maxEqSize)

    Conjunction(quans, arithConj, negConjs, to)
  }
  
  private def randomNegConjs(sizeFactor : Double,
                             maxEqNum : Int, maxEqSize : Int) : NegatedConjunctions = {
    val conjs = for (_ <- PlainRange(0, Debug.random(0, Math.max(1, sizeFactor.toInt))))
                yield randomConj(sizeFactor, maxEqNum, maxEqSize)
    
    NegatedConjunctions(conjs, to)
  }
  
  /**
   * Generate arbitrary instances of <code>ArithConj</code>
   */
  def testArithConj1 = {
    for (eqNumConj <- PlainRange(5); eqNumDisj <- PlainRange(5); _ <- PlainRange(10)) {
      val inputConj = (for (len <- Debug.randoms(0, 8)) yield randomLC(len))
                      .take(eqNumConj).toList
      val eqConj = EquationConj(inputConj, to)
      val negInputConj = (for (len <- Debug.randoms(0, 8)) yield randomLC(len))
                         .take(eqNumDisj).toList
      val negEqConj = NegEquationConj(negInputConj, to)
      
      val ac = ArithConj(eqConj, negEqConj, to)
      val ac2 = ArithConj.conj((for (lc <- inputConj) yield EquationConj(lc, to)) :::
                               (for (lc <- negInputConj) yield NegEquationConj(lc, to)),
                               to)

      assertEquals(ac, ac2)
                               
      ac.sortBy(toRev)
      ac.internallyPropagate
    }
  }

  /**
   * Generate arbitrary instances of <code>ArithConj</code>
   */
  def testArithConj2 = {
    for (eqNumConj <- PlainRange(5); eqNumDisj <- PlainRange(5); _ <- PlainRange(10)) {
      val ac = randomAC(eqNumConj, eqNumDisj, 8)

      ac.sortBy(toRev)
      ac.internallyPropagate
    }
  }

  /**
   * Generate arbitrary instances of <code>Conjunction</code> and perform
   * various operations on them
   */
  def testConj1 = {
    for (sizeFactor <- PlainRange(0, 10); _ <- PlainRange(100)) {
      val conj = randomConj(sizeFactor, 5, 8)
      conj.constants
      conj.variables
      
//      println(conj)
      
      // test what happens when the free variables are quantified
      var quantConj = conj
//      println("Quantifying ...")
      while (!quantConj.variables.isEmpty) {
        val newQuantConj =
          Conjunction.quantify(Array(Quantifier.ALL), quantConj, to)
//          println(newQuantConj)
        assertEquals(newQuantConj.constants, conj.constants)
        assertEquals(newQuantConj.variables,
                     for (VariableTerm(i) <- quantConj.variables; if (i >= 1))
                     yield VariableTerm(i - 1))
	quantConj = newQuantConj
      }

      // test to change the ordering of a conjunction
      val conjRev = conj.sortBy(toRev)
      conjRev.constants
      conjRev.variables
      
      // test negation
      assertEquals(conj.negate.negate, conj)
      
      // test conjunction/disjunction with neutral elements
      assertEquals(Conjunction.conj(Array(conj, Conjunction.TRUE), to), conj)
      assertEquals(Conjunction.conj(Array(conj, NegatedConjunctions.TRUE), to), conj)
      assertEquals(Conjunction.conj(Array(conj, ArithConj.TRUE), to), conj)
      assertEquals(Conjunction.conj(Array(conj, EquationConj.TRUE), to), conj)
      assertEquals(Conjunction.disj(Array(conj, Conjunction.FALSE), to), conj)
      assertEquals(Conjunction.disj(Array(conj, NegatedConjunctions.FALSE), to), conj)
      assertEquals(Conjunction.disj(Array(conj, ArithConj.FALSE), to), conj)
      assertEquals(Conjunction.disj(Array(conj, EquationConj.FALSE), to), conj)
      
      // test idempotence
      assertEquals(Conjunction.conj(Array(conj, conj), to), conj)
      assertEquals(Conjunction.disj(Array(conj, conj), to), conj)
    }
  }
}
