package ap.terfor.equations;

import ap.terfor.linearcombination.{LinearCombination, LCBlender}
import ap.basetypes.IdealInt
import ap.util.{Debug, Logic}

/**
 * Reduce a term (currently: a linear combination) by rewriting with equations.
 * The equations have to be given in form of a mapping from atomic terms
 * (constants or variables) to linear combinations
 */
object ReduceWithEqs {

  private val AC = Debug.AC_EQUATIONS

  def apply(eqs : scala.collection.Map[Term, LinearCombination], order : TermOrder)
                                                  : ReduceWithEqs =
    new ReduceWithEqs (eqs, order)

  def apply(eqs : EquationConj, order : TermOrder)
                                                  : ReduceWithEqs = {
    ////////////////////////////////////////////////////////////////////////////
    Debug.assertPre(AC, eqs isSortedBy order)
    ////////////////////////////////////////////////////////////////////////////
    ReduceWithEqs(eqs.toMap, order)
  }
}

/**
 * Reduce a term (currently: a linear combination) by rewriting with equations.
 * The equations have to be given in form of a mapping from atomic terms
 * (constants or variables) to linear combinations
 */
class ReduceWithEqs private (private val equations
                                : scala.collection.Map[Term, LinearCombination],
                             private val order : TermOrder)
                    extends ((TerFor) => TerFor) {

  def apply(lc : LinearCombination) : LinearCombination = {
    ////////////////////////////////////////////////////////////////////////////
    Debug.assertPre(ReduceWithEqs.AC, lc isSortedBy order)
    ////////////////////////////////////////////////////////////////////////////

    val blender = new LCBlender (order)
    blender += (IdealInt.ONE, lc)
    val changed = runBlender(blender)
    
    if (changed) blender.result else lc
  }

  /**
   * Run the <code>blender</code> and add linear combinations from
   * <code>eqs</code> whenever it is possible to reduce monomials.
   * <code>true</code> is returned in case any reduction was performed
   */
  private def runBlender(blender : LCBlender) : Boolean = {
    var changed : Boolean = false
    while (blender.hasNext) {
      val (nextCoeff, nextTerm) = blender.peekNext
                 
      (equations get nextTerm) match {
      case Some(eq) => {
        ////////////////////////////////////////////////////////////////////////
        Debug.assertPre(ReduceWithEqs.AC,
                        eq.leadingTerm == nextTerm && (eq isSortedBy order))
        ////////////////////////////////////////////////////////////////////////
                       
        val (quot, rem) = (nextCoeff reduceAbs eq.leadingCoeff)
        if (rem != nextCoeff) {
          blender += (-quot, eq)
          changed = true
        }
                   
        if (blender.hasNext && (blender.peekNext _2) == nextTerm) blender.next
      }

      case None => blender.next
      }
    }
    
    changed
  }
  
  /**
   * Same as <code>apply(lc:LinearCombination)</code>, but also multiply
   * <cocde>lc</code> with integers in case this allows to eliminate the leading
   * term (pseudo-division). It is ensured that the resulting
   * <code>LinearCombination</code> has a positive leading coefficient
   */
  private def pseudoReduce(lc : LinearCombination) : LinearCombination = {
    var curLC = apply(lc)
    
    while (!curLC.isZero && (equations contains curLC.leadingTerm)) {
      val eq = equations(curLC.leadingTerm)
      val curLCCoeff = curLC.leadingCoeff
      val eqCoeff = eq.leadingCoeff
      
      //////////////////////////////////////////////////////////////////////////
      Debug.assertInt(ReduceWithEqs.AC,
                      eq.leadingTerm == curLC.leadingTerm &&
                      !(eqCoeff divides curLCCoeff))
      //////////////////////////////////////////////////////////////////////////

      val lcGcd = curLCCoeff gcd eqCoeff
      
      val blender = new LCBlender (order)
      blender ++= Array((eqCoeff / lcGcd, curLC), (-curLCCoeff / lcGcd, eq))
      runBlender(blender)
                        
      //////////////////////////////////////////////////////////////////////////
      Debug.assertInt(ReduceWithEqs.AC,
                      { val res = blender.result
                        res.isZero ||
                          order.compare(res.leadingTerm, curLC.leadingTerm) < 0 })
      //////////////////////////////////////////////////////////////////////////

      curLC = blender.result
    }
    
    if (curLC.isZero) {
      LinearCombination.ZERO
    } else if (curLC.leadingCoeff.signum < 0) {
      // when the leading coefficient of the <code>LinearCombination</code> is
      // made positive, it might be possible to apply further reductions
      val blender = new LCBlender (order)
      blender += (IdealInt.MINUS_ONE, curLC)
      runBlender(blender)
      blender.result
    } else {
      curLC
    }
  }

  def apply(conj : NegEquationConj) : NegEquationConj = {
    ////////////////////////////////////////////////////////////////////////////
    Debug.assertPre(ReduceWithEqs.AC, conj isSortedBy order)
    ////////////////////////////////////////////////////////////////////////////
    
    val res = NegEquationConj(for (lc <- conj.elements)
                              yield pseudoReduce(lc), order)
    ////////////////////////////////////////////////////////////////////////////
    Debug.assertPost(ReduceWithEqs.AC,
      res.isFalse ||
      Logic.forall(for (lc <- res.elements) yield
                   (!(equations contains lc.leadingTerm) &&
                    Logic.forall(for ((c, t) <- lc.elements) yield (
                                   equations.get(t) match {
                                   case Some(eq) => c isAbsMinMod eq.leadingCoeff
                                   case None => true
                                   })))))
    ////////////////////////////////////////////////////////////////////////////
    res
  }
  
  /**
   * The dispatcher for the different term and formula classes
   */
  def apply(toBeReduced : TerFor) : TerFor = toBeReduced match {
    case lc : LinearCombination => apply(lc)
    case conj : NegEquationConj => apply(conj)
  }
}